import sys
import types
import typing as tp

from satella.coding.typing import ExceptionClassType
from satella.instrumentation import Traceback
from satella.opentracing.trace import Span


def trace_exception(span: tp.Optional[Span], exc_type: tp.Optional[ExceptionClassType] = None,
                    exc_val: tp.Optional[Exception] = None,
                    exc_tb: tp.Optional[types.TracebackType] = None) -> None:
    """
    Log an exception's information to the chosen span, as logs and tags.

    :param span: span to log into or None for a no-op
    :param exc_type: exception type. If None this will be taken from sys.exc_info.
    :param exc_val: exception value. If None this will be taken from sys.exc_info.
    :param exc_tb: exception traceback. If None this will be taken from sys.exc_info.
    """
    if span is None:
        return

    if exc_type is None:
        exc_type, exc_val, exc_tb = sys.exc_info()
        if exc_type is None:
            return

    span.set_tag('error', True)
    span.log_kv({'event': 'error',
                 'message': str(exc_val),
                 'error.object': exc_val,
                 'error.kind': exc_type,
                 'stack': Traceback(exc_tb.tb_frame).pretty_format()
                 })



import math
import typing as tp
from .monitor import Monitor
from ...exceptions import AlreadyAllocated, Empty


class SequentialIssuer(Monitor):
    """
    A classs that issues an monotonically increasing value.

    :param start: start issuing IDs from this value

    :ivar start: next value to be issued
    """
    __slots__ = ('start',)

    def __init__(self, start: int = 0):
        super().__init__()
        self.start = start

    @Monitor.synchronized
    def issue(self) -> int:
        """
        Just issue a next identifier

        :return: a next identifier
        """
        try:
            return self.start
        finally:
            self.start += 1

    @Monitor.synchronized
    def no_less_than(self, no_less_than: int) -> int:
        """
        Issue an int, which is no less than a given value

        :param no_less_than: value that the returned id will not be less than this
        :return: an identifier, no less than no_less_than
        """
        try:
            if self.start > no_less_than:
                return self.start
                self.start += 1
            else:
                self.start = no_less_than
                return self.start
        finally:
            self.start += 1


class IDAllocator(Monitor):
    """
    Reusable ID allocator

    You can use it to requisition ints from a pool, and then free their ints, permitting
    them to be reused.

    Thread-safe.

    :param start_at: the lowest integer that the allocator will return
    :param top_limit: the maximum value that will not be allocated. If used,
        subsequent calls to :meth:`~satella.coding.concurrent.IDAllocator.allocate_int` will
        raise :class:`~satella.exceptions.Empty`
    """
    __slots__ = 'start_at', 'ints_allocated', 'free_ints', 'bound', 'top_limit'

    def __init__(self, start_at: int = 0, top_limit: tp.Optional[int] = None):
        super().__init__()
        self.start_at = start_at
        self.ints_allocated = set()
        self.free_ints = set()
        self.top_limit = top_limit or math.inf
        self.bound = 0

    def _extend_the_bound_to(self, x: int) -> int:
        """Return how many integers added"""
        if x > self.top_limit:
            x = self.top_limit
        how_many = 0
        for i in range(self.bound, x):
            self.free_ints.add(i)
            how_many += 1
        self.bound = x
        return how_many

    @Monitor.synchronized
    def mark_as_free(self, x: int):
        """
        Mark x as free

        :param x: int to free
        :raises ValueError: x was not allocated or less than start_at
        """
        if x < self.start_at:
            raise ValueError('%s is less than start_at' % (x,))
        x -= self.start_at
        if x not in self.ints_allocated:
            raise ValueError('%s was not allocated' % (x,))
        self.ints_allocated.remove(x)
        self.free_ints.add(x)

    @Monitor.synchronized
    def allocate_int(self) -> int:
        """
        Return a previously unallocated int, and mark it as allocated

        :return: an allocated int
        :raises Empty: could not allocate an int due to top limit
        """
        if not self.free_ints:
            if self._extend_the_bound_to(self.bound + 10) == 0:
                raise Empty('No integers remaining!')
        x = self.free_ints.pop()
        self.ints_allocated.add(x)
        return x + self.start_at

    @Monitor.synchronized
    def mark_as_allocated(self, x: int):
        """
        Mark given x as allocated

        :param x: x to mark as allocated
        :raises AlreadyAllocated: x was already allocated
        :raises ValueError: x is less than start_at
        """
        if x < self.start_at:
            raise ValueError('%s is less than start_at' % (x,))
        if x >= self.top_limit:
            raise ValueError('Cannot allocate a value greater or equal than top limit!')
        x -= self.start_at
        if x >= self.bound:
            self._extend_the_bound_to(x + 1)
            self.free_ints.remove(x)
            self.ints_allocated.add(x)
        else:
            if x in self.ints_allocated:
                raise AlreadyAllocated()
            self.free_ints.remove(x)
            self.ints_allocated.add(x)



import logging

logger = logging.getLogger(__name__)

from satella.imports import import_from

__all__ = []


def do_import():
    import_from(__path__, __name__, __all__, locals(), recursive=True, fail_on_attributerror=False,
                create_all=True)



import inspect
import io

try:
    # noinspection PyPep8Naming
    import cPickle as pickle
except ImportError:
    import pickle
import sys
import traceback
import types
import typing as tp

from satella.json import JSONAble
from .classes import StackFrame, GenerationPolicy


def frame_from_traceback(tb: types.TracebackType) -> types.FrameType:
    """
    Extract the bottom stack frame from a traceback

    :param tb: traceback to extract the frame
    :return: bottom stack frame
    """
    while tb.tb_next:
        tb = tb.tb_next
    return tb.tb_frame


class Traceback(JSONAble):
    """
    Class used to preserve exceptions and chains of stack frames.
     Picklable.

    If starting frame is not given, an exception must be in progress.

    You can also convert it to a secure representation, ie. one that will be completely JSON
    and thus safe to load from untrusted sources. You will not lose the unpicklability of them by doing
    so, as they can be safely reconstructed (pickle will be encoded as base64 string).

    :param starting_frame: frame to start tracking the traceback from.
        Must be either None, in which case an exception must be in progress and will be taken
        else must be an instance of <class 'frame'>.
    :param policy: policy for traceback generation
    :raise ValueError: there is no traceback to get info from!
        Is any exception in process?
     """
    __slots__ = ('formatted_traceback', 'frames')

    def __str__(self) -> str:
        out = ['Traceback:\n']
        for frame in self.frames:
            out.append(str(frame))
        return ''.join(out)

    def __init__(self, starting_frame: tp.Optional[types.FrameType] = None,
                 policy=GenerationPolicy):
        if inspect.isclass(policy):
            value_pickling_policy = policy()
        else:
            value_pickling_policy = policy

        self.frames = []  # type: tp.List[StackFrame]

        if starting_frame is None:
            tb = sys.exc_info()[2]
            if tb is not None:
                f = frame_from_traceback(tb)
            else:
                f = inspect.currentframe()
        else:
            f = starting_frame

        while f:
            self.frames.append(StackFrame(f, value_pickling_policy))
            f = f.f_back

        self.formatted_traceback = str(traceback.format_exc())

    def __eq__(self, other: 'Traceback') -> bool:
        return self.frames == other.frames

    def pickle_to(self, stream: tp.BinaryIO) -> None:
        """Pickle self to target stream"""
        pickle.dump(self, stream, pickle.HIGHEST_PROTOCOL)

    def pickle(self) -> bytes:
        """Returns this instance, pickled"""
        bio = io.BytesIO()
        self.pickle_to(bio)
        return bio.getvalue()

    @classmethod
    def from_pickle(cls, pick: tp.Union[io.BytesIO, bytes]) -> 'Traceback':
        """
        Load a traceback from a pickle

        :param pick: either bytes or a BytesIO to load it from
        :return: previously serialized Traceback
        :raises ValueError: unserialized object is not a Traceback!
        """
        if isinstance(pick, io.BytesIO):
            a = pickle.load(pick)
        else:
            a = pickle.loads(pick)
        if not isinstance(a, Traceback):
            raise ValueError('%s is not a traceback!' % (type(a),))
        return a

    def to_json(self) -> dict:
        return {
            'frames': [frame.to_json() for frame in self.frames],
            'formatted_traceback': self.formatted_traceback
        }

    @classmethod
    def from_json(cls, x: dict) -> 'Traceback':
        tb = Traceback.__new__(Traceback)
        tb.frames = [StackFrame.from_json(y) for y in x['frames']]
        tb.formatted_traceback = x['formatted_traceback']
        return tb

    def pretty_format(self) -> str:
        """
        Return a multi-line, pretty-printed representation of all exception
        data.

        :return: text
        """
        bio = io.StringIO()
        self.pretty_print(bio)
        return bio.getvalue()

    def pretty_print(self, output: tp.TextIO = sys.stderr) -> None:
        """
        Pretty-print the exception

        :param output: a file-like object in text mode
        :return: unicode
        """
        output.write(self.formatted_traceback)
        output.write(u'\n* Stack trace, innermost first\n')
        for frame in self.frames:
            output.write(u'** %s at %s:%s\n' % (frame.name, frame.filename, frame.lineno))
            for name, value in frame.locals.items():
                try:
                    output.write(u'*** %s: %s\n' % (name, value.repr))
                except BaseException as e:
                    output.write(
                        u'*** %s: repr unavailable (due to locally raised %s)\n' % (name, repr(e)))



from satella.coding.typing import Number


class Vector(tuple):
    """
    A tuple that allows place-wise addition and subtraction of it's entries, and also
    tuple-times-float operations

    Ie.

    >>> a = Vector((2, 3))
    >>> b = Vector((1, 2))
    >>> assert a-b == Vector((1, 1))
    >>> assert a*2 == Vector((4, 6))
    """

    def __add__(self, other: 'Vector') -> 'Vector':
        result = []
        for a, b in zip(self, other):
            result.append(a + b)
        return Vector(result)

    def __sub__(self, other: 'Vector') -> 'Vector':
        result = []
        for a, b in zip(self, other):
            result.append(a - b)
        return Vector(result)

    __iadd__ = __add__
    __isub__ = __sub__

    def __mul__(self, other: Number) -> 'Vector':
        result = []
        for a in self:
            result.append(a * other)
        return Vector(result)

    def __truediv__(self, other: Number) -> 'Vector':
        result = []
        for a in self:
            result.append(a / other)
        return Vector(result)

    __imul__ = __mul__
    __itruediv__ = __truediv__



from .measure import measure, TimeSignal
from .misc import time_us, time_ms, time_as_int, sleep
from .parse import parse_time_string
from .backoff import ExponentialBackoff

__all__ = ['measure', 'TimeSignal', 'ExponentialBackoff', 'time_us', 'time_ms',
           'time_as_int', 'parse_time_string', 'sleep']



import time
import typing as tp

from satella.coding.decorators.decorators import wraps

from satella.coding.typing import ExceptionList

from satella.coding.concurrent.thread import Condition
from .measure import measure
from ..exceptions import WouldWaitMore


class ExponentialBackoff:
    """
    A class that will sleep increasingly longer on errors. Meant to be used in such a way:

    >>> eb = ExponentialBackoff(start=2, limit=30)
    >>> while not connect():
    >>>     eb.failed()
    >>>     eb.sleep()
    >>> eb.success()

    Also a structure that will mark an object (eg. the Internet access) as inaccessible for some
    duration. Usage is that case is like this:

    >>> eb = ExponentialBackoff(start=2, limit=30)
    >>> eb.failed()
    >>> self.assertFalse(eb.available)
    >>> time.sleep(2)
    >>> self.assertTrue(eb.available)

    Note that this structure is thread safe only when a single object is doing
    the :code:`success` or :code:`failed` calls, and other utilize
    :meth:`~satella.time.ExponentialBackoff.wait_until_available`.

    :param start: value at which to start
    :param limit: maximum sleep timeout
    :param sleep_fun: function used to sleep. Will accept a single argument - number of
        seconds to wait
    :param grace_amount: amount of fails() that this will survive before everything fails
    """
    __slots__ = 'start', 'limit', 'counter', 'sleep_fun', 'unavailable_until', 'condition', \
        'grace_amount', 'grace_counter'

    def __init__(self, start: float = 1, limit: float = 30,
                 sleep_fun: tp.Callable[[float], None] = time.sleep,
                 grace_amount: int = 0):
        self.start = start
        self.grace_amount = grace_amount
        self.grace_counter = 0
        self.limit = limit
        self.counter = 0
        self.sleep_fun = sleep_fun
        self.condition = Condition()
        self.unavailable_until = None

    def sleep(self) -> None:
        """
        Called when sleep is expected.
        """
        self.sleep_fun(self.counter)

    def failed(self) -> None:
        """
        Called when something fails.
        """
        if self.grace_amount == self.grace_counter:
            if self.counter == 0:
                self.counter = self.start
            else:
                self.counter = min(self.limit, self.counter * 2)
            self.unavailable_until = time.monotonic() + self.counter
        else:
            self.grace_counter += 1

    def wait_until_available(self, timeout: tp.Optional[float] = None) -> None:
        """
        Waits until the service is available

        :param timeout: maximum amount of seconds to wait. If waited more than that,
            WouldWaitMore will be raised
        :raises WouldWaitMore: waited for timeout and service still was not healthy
        """
        with measure(timeout=timeout) as m:
            while not m.timeouted:
                tn = self.time_until_next_check()
                if tn is None:
                    return
                self.condition.wait(timeout=tn, dont_raise=True)
            raise WouldWaitMore('timeouted while waiting for service to become healthy')

    def time_until_next_check(self) -> tp.Optional[float]:
        """Return the time until next health check, or None if the service is healthy"""
        if self.unavailable_until is None:
            return None
        else:
            t = time.monotonic()
            if t > self.unavailable_until:
                self.unavailable_until = None
                return None
            else:
                return self.unavailable_until - t

    @property
    def ready_for_next_check(self) -> bool:
        """
        :return: Has  :meth:`~satella.time.ExponentialBackoff.failure` been called only
            in the last waiting period?
        """
        if self.unavailable_until is None:
            return True
        elif time.monotonic() > self.unavailable_until:
            self.unavailable_until = None
            return True
        else:
            return False

    @property
    def available(self) -> bool:
        """Was the status of the last call :code:`success`?"""
        return self.counter == 0

    def success(self) -> None:
        """
        Called when something successes.
        """
        self.counter = 0
        self.grace_counter = 0
        self.unavailable_until = None
        self.condition.notify_all()

    def launch(self, exceptions_on_failed: ExceptionList = Exception):
        """
        A decorator to simplify writing doing-something loops. Basically, this:

        >>> eb = ExponentialBackoff(start=2.5, limit=30)
        >>> @eb.launch(TypeError)
        >>> def do_action(*args, **kwargs):
        >>>     x_do_action(*args, **kwargs)
        >>> do_action(5, test=True)

        is equivalent to this:

        >>> eb = ExponentialBackoff(start=2.5, limit=30)
        >>> while True:
        >>>     try:
        >>>         x_do_action(5, test=True)
        >>>     except TypeError:
        >>>         eb.failed()
        >>>         eb.sleep()

        :param exceptions_on_failed: a list of a single exception of exceptions
            whose raising will signal that fun has failed
        :return: a function, that called, will pass the exactly same parameters
        """
        def outer(fun):
            @wraps(fun)
            def inner(*args, **kwargs):
                try:
                    r = fun(*args, **kwargs)
                    self.success()
                    return r
                except exceptions_on_failed:
                    self.failed()
                    self.sleep()
            return inner
        return outer



import logging
import unittest

from satella.instrumentation.metrics.data import MetricData, MetricDataCollection

logger = logging.getLogger(__name__)


class TestMetricData(unittest.TestCase):

    def test_internal(self):
        md = MetricData('name', 2, {'labels': 'key'}, None, True)
        md2 = MetricData.from_json(md.to_json())
        self.assertTrue(md2.internal)

    def test_metric_data_collection_add(self):
        a = MetricDataCollection(MetricData('root', 3, {'labels': 'key'}),
                                 MetricData('root_a', 3, {'labels': 'key'}))
        a += MetricDataCollection(MetricData('root', 2, {'labels': 'key'}),
                                  MetricData('root_a', 4, {'labels': 'key'}))

    def test_update_labels_2(self):
        a = MetricDataCollection(MetricData('root', 2, {'labels': 'key'}))
        a.add_labels({'service': 'wtf'})
        self.assertEqual(next(iter(a.values)).labels, {'labels': 'key', 'service': 'wtf'})

    def test_json_serialization(self):
        a = MetricDataCollection(MetricData('root', 2, {'labels': 'key'}))
        b = a.to_json()
        self.assertTrue(a.strict_eq(MetricDataCollection.from_json(b)))

    def test_update_labels(self):
        a = MetricData('root', 2, {'labels': 'key'})
        a.add_labels({'service': 'wtf'})
        self.assertEqual(a.labels, {'labels': 'key', 'service': 'wtf'})

    def test_update_3(self):
        a1 = MetricDataCollection(MetricData('root.metric', 2, {'service': 'my_service'}))
        a1 += MetricData('root.metric', 10, {'service': 'my_service'}, 10)
        self.assertTrue(MetricDataCollection(
            MetricData('root.metric', 10, {'service': 'my_service'}, 10)).strict_eq(a1))

    def test_update_4(self):
        a1 = MetricDataCollection(
            MetricData('root.metric', 2, {'service': 'my_service', 'slot': 1}))
        a1 += MetricDataCollection(
            MetricData('root.metric', 10, {'service': 'my_service', 'slot': 1}, 10))
        self.assertTrue(MetricDataCollection(
            MetricData('root.metric', 10, {'service': 'my_service', 'slot': 1}, 10)).strict_eq(a1))

    def test_update(self):
        a1 = MetricDataCollection(MetricData('root.metric', 25.0, {'period': 1}),
                                  MetricData('root.metric', 50.0, {'period': 2}))
        a2 = MetricDataCollection(MetricData('root.metric', 20.0, {'period': 1}))
        a3 = a1 + a2
        self.assertTrue(MetricDataCollection(MetricData('root.metric', 20.0, {'period': 1}),
                                             MetricData('root.metric', 50.0,
                                                        {'period': 2})).strict_eq(a3))

    def test_postfix_and_prefix(self):
        a = MetricDataCollection(MetricData('root', 3))
        a.prefix_with('test')
        self.assertTrue(MetricDataCollection(MetricData('test.root', 3)).strict_eq(a))

        a = MetricDataCollection(MetricData('root', 3))
        a.postfix_with('test')
        self.assertTrue(MetricDataCollection(MetricData('root.test', 3)).strict_eq(a))

    def test_update_2(self):
        a = MetricDataCollection(MetricData('root', 3, {'a': 5}),
                                 MetricData('root.sum', 3, {'a': 5}))
        b = MetricDataCollection(MetricData('root', 7, {'a': 5}),
                                 MetricData('root.sum', 8, {'a': 3}))
        a += b
        self.assertTrue(MetricDataCollection(MetricData('root', 7, {'a': 5}),
                                             MetricData('root.sum', 8, {'a': 3}),
                                             MetricData('root.sum', 3, {'a': 5})).strict_eq(a))



from .atomic import AtomicNumber
from .callablegroup import CallableGroup, CallNoOftenThan, CancellableCallback
from .functions import parallel_execute, run_as_future
from .futures import Future, WrappingFuture, InvalidStateError, FutureCollection
from .id_allocator import IDAllocator, SequentialIssuer
from .locked_dataset import LockedDataset
from .locked_structure import LockedStructure
from .monitor import MonitorList, Monitor, MonitorDict, RMonitor, MonitorSet
from .sync import sync_threadpool
from .thread import TerminableThread, Condition, SingleStartThread, call_in_separate_thread, \
    BogusTerminableThread, IntervalTerminableThread
from .timer import Timer
from .thread_collection import ThreadCollection
from .queue import PeekableQueue

__all__ = ['LockedDataset', 'Monitor', 'RMonitor', 'CallableGroup', 'TerminableThread',
           'MonitorDict', 'MonitorList', 'Condition', 'LockedStructure', 'AtomicNumber',
           'CallNoOftenThan', 'SingleStartThread', 'IDAllocator', 'call_in_separate_thread',
           'BogusTerminableThread', 'Timer', 'parallel_execute', 'run_as_future',
           'sync_threadpool', 'IntervalTerminableThread', 'Future', 'MonitorSet',
           'WrappingFuture', 'InvalidStateError', 'PeekableQueue',
           'CancellableCallback', 'ThreadCollection', 'FutureCollection',
           'SequentialIssuer']



import unittest

from satella.coding.sequences import n_th
from satella.instrumentation.metrics import getMetric

import time
from satella.instrumentation.metrics.structures import MetrifiedThreadPoolExecutor, \
    MetrifiedCacheDict, MetrifiedLRUCacheDict, MetrifiedExclusiveWritebackCache
from .test_metrics import choose


def wait():
    time.sleep(2)


class TestThreadPoolExecutor(unittest.TestCase):
    def test_exclusive_writeback_cache(self):
        cache_hits = getMetric('wbc.cachedict.hits', 'counter')
        cache_miss = getMetric('wbc.cachedict.miss', 'counter')
        entries_waiting = getMetric('wbc.cachedict.waiting', 'callable')
        a = {5: 3, 4: 2, 1: 0}
        b = {'no_calls': 0}

        def setitem(k, v):
            nonlocal a, b
            b['no_calls'] += 1
            a[k] = v

        def getitem(k):
            nonlocal a, b
            b['no_calls'] += 1
            return a[k]

        def delitem(k):
            nonlocal a, b
            b['no_calls'] += 1
            del a[k]

        wbc = MetrifiedExclusiveWritebackCache(setitem, getitem, delitem,
                                               cache_hits=cache_hits,
                                               cache_miss=cache_miss,
                                               entries_waiting=entries_waiting)
        self.assertEqual(wbc[5], 3)
        self.assertEqual(b['no_calls'], 1)
        self.assertRaises(KeyError, lambda: wbc[-1])
        self.assertEqual(b['no_calls'], 2)
        self.assertEqual(wbc[5], 3)
        self.assertEqual(b['no_calls'], 2)
        self.assertEqual(n_th(cache_miss.to_metric_data().values).value, 2)
        self.assertEqual(n_th(cache_hits.to_metric_data().values).value, 1)
        wbc[5] = 2
        wbc.sync()
        self.assertEqual(a[5], 2)
        self.assertEqual(b['no_calls'], 3)
        del wbc[4]
        wbc.sync()
        self.assertRaises(KeyError, lambda: a[4])

    def test_metrified_cache_dict(self):
        cache_hits = getMetric('cachedict.hits', 'counter')
        cache_miss = getMetric('cachedict.miss', 'counter')
        refreshes = getMetric('refreshes', 'counter')
        how_long_takes = getMetric('how_long_takes', 'summary')
        value = 2

        def getter(key):
            nonlocal value
            time.sleep(0.5)
            if value is None:
                raise KeyError()
            else:
                return value

        mcd = MetrifiedCacheDict(1, 2, getter, cache_hits=cache_hits,
                                 cache_miss=cache_miss,
                                 refreshes=refreshes,
                                 how_long_refresh_takes=how_long_takes)
        mcd[2]
        self.assertEqual(n_th(cache_hits.to_metric_data().values).value, 0)
        self.assertEqual(n_th(cache_miss.to_metric_data().values).value, 1)
        self.assertEqual(n_th(refreshes.to_metric_data().values).value, 1)
        self.assertGreaterEqual(n_th(how_long_takes.to_metric_data().values).value, 0.5)
        mcd[2]
        self.assertEqual(n_th(cache_hits.to_metric_data().values).value, 1)
        self.assertEqual(n_th(cache_miss.to_metric_data().values).value, 1)
        self.assertEqual(n_th(refreshes.to_metric_data().values).value, 1)

    def test_metrified_lru_cache_dict(self):
        cache_hits = getMetric('lrucachedict.hits', 'counter')
        cache_miss = getMetric('lrucachedict.miss', 'counter')
        refreshes = getMetric('lrurefreshes', 'counter')
        evictions = getMetric('lruevictions', 'counter')
        how_long_takes = getMetric('lruhow_long_takes', 'summary')
        value = 2

        def getter(key):
            nonlocal value
            time.sleep(0.5)
            if value is None:
                raise KeyError()
            else:
                return value

        mcd = MetrifiedLRUCacheDict(1, 2, getter, cache_hits=cache_hits,
                                 cache_miss=cache_miss,
                                 refreshes=refreshes,
                                 how_long_refresh_takes=how_long_takes,
                                 evictions=evictions,
                                 max_size=2)
        mcd[2]
        self.assertEqual(n_th(cache_hits.to_metric_data().values).value, 0)
        self.assertEqual(n_th(cache_miss.to_metric_data().values).value, 1)
        self.assertEqual(n_th(refreshes.to_metric_data().values).value, 1)
        self.assertGreaterEqual(n_th(how_long_takes.to_metric_data().values).value, 0.5)
        mcd[2]
        self.assertEqual(n_th(cache_hits.to_metric_data().values).value, 1)
        self.assertEqual(n_th(cache_miss.to_metric_data().values).value, 1)
        self.assertEqual(n_th(refreshes.to_metric_data().values).value, 1)
        mcd[3]
        self.assertEqual(n_th(evictions.to_metric_data().values).value, 0)
        mcd[4]
        self.assertEqual(n_th(evictions.to_metric_data().values).value, 1)

    def test_metrified_thread_pool_executor(self):
        waiting_summary = getMetric('mtpe.summary', 'summary')
        executing_summary = getMetric('mtpe.executing', 'summary')
        callable_metric = getMetric('mtpe.outstanding', 'callable')

        mtpe = MetrifiedThreadPoolExecutor(max_workers=2, time_spent_waiting=waiting_summary,
                                           time_spent_executing=executing_summary,
                                           waiting_tasks=callable_metric)

        mtpe.submit(wait)
        time.sleep(0.1)
        self.assertEqual(n_th(callable_metric.to_metric_data().values).value, 0)
        mtpe.submit(wait)
        fr = mtpe.submit(wait)
        time.sleep(0.1)
        self.assertEqual(n_th(callable_metric.to_metric_data().values).value, 1)
        fr.result()
        self.assertIn(choose('.count', executing_summary.to_metric_data()).value, {2, 3})
        self.assertEqual(choose('.count', waiting_summary.to_metric_data()).value, 3)



from .collector import calculate_occupancy_factor, sleep_cpu_aware, CPUTimeManager
from .concurrency import CPUTimeAwareIntervalTerminableThread

__all__ = ['calculate_occupancy_factor', 'sleep_cpu_aware',
           'CPUTimeAwareIntervalTerminableThread',
           'CPUTimeManager']



from ..exceptions import WouldWaitMore
import typing as tp
import time
import copy
import inspect
import warnings
from concurrent.futures import Future
from functools import wraps  # import from functools to prevent circular import exception


TimeSignal = tp.Callable[[], float]


class measure:
    """
    A class used to measure time elapsed. Use for example like this:

    >>> with measure() as measurement:
    >>>     time.sleep(1)
    >>>     print('This has taken so far', measurement(), 'seconds')
    >>>     time.sleep(1)
    >>> print('A total of ', measurement(), 'seconds have elapsed')

    You can also use the .start() method instead of context manager. Time measurement
    will stop after exiting or calling .stop() depending on stop_on_stop flag.

    Time elapsing starts after the counter is created. If you wish to start it manually,
    please specify create_stopped=True

    When instantiated and called with no arguments, this will return the time elapsed:

    >>> a = measure()
    >>> time.sleep(0.5)
    >>> assert a() >= 0.5

    You can also decorate your functions to have them keep track time of their execution, like that:

    >>> @measure()
    >>> def measuring(measurement_object: measure, *args, **kwargs):
    >>>     ...

    This will also correctly work on methods, correctly inserting the measurement object after
    self/cls:

    >>> class Test:
    >>>     @measure()
    >>>     def measuring(self, measurement_object: measure):
    >>>         ...

    You can also measure how long does executing a future take, eg.

    >>> future = get_my_future()
    >>> measurement = measure(future)
    >>> future.result()
    >>> print('Executing the future took', measurement(), 'seconds')

    In case a future is passed, the measurement will stop automatically as soon as the future
    returns with a result (or exception).

    Note that in order to reuse a single counter you must .reset() it first. Just calling .start()
    after .stop() will result in the timer being resumed instead!

    Note that if you're measuring generators, this will count the time passed from the measure()
    was called till the generator finishes.

    Note that if you're using the decorator form, this object will be first copied and then
    passed to the function/future. This is to prevent undefined behaviour during multithreading.
    Also, the timer that you pass to this function will be started/not started, depending on what
    you set earlier. .reset() will be called on a copy of this object.

    This can also be used to write custom timeouts, eg.

    >>> with measure(timeout=5) as m:
    >>>     while not m.timeouted:
    >>>         ... do something ...
    >>>         if condition:
    >>>             break
    >>>     if m.timeouted:
    >>>         raise WouldWaitMore('timeout hit')

    :param stop_on_stop: stop elapsing time upon calling .stop()/exiting the context manager.
        If this is set to False then .start() and .stop() won't work and calling them will raise a
        TypeError.
    :param adjust: interval to add to current time upon initialization
    :param time_getter_callable: callable/0 -> float to get the time with
    :param create_stopped: if this is set to True, you will manually need to call .start()
    :param timeout: a time limit, after exceeding which the property `timeouted` will be true
    """
    __slots__ = ('started_on', 'elapsed', 'stopped_on', 'stop_on_stop', 'time_getter_callable',
                 'create_stopped', 'timeout')

    def __init__(self, future_to_measure: tp.Optional[Future] = None, stop_on_stop: bool = True,
                 adjust: float = 0.0, time_getter_callable: TimeSignal = time.monotonic,
                 create_stopped: bool = False,
                 timeout: tp.Optional[float] = None):
        self.time_getter_callable = time_getter_callable
        self.timeout = timeout
        self.started_on = time_getter_callable() + adjust
        self.create_stopped = create_stopped
        if create_stopped:
            self.elapsed = 0
            self.stopped_on = self.started_on
        else:
            self.elapsed = None
            self.stopped_on = None
        self.stop_on_stop = stop_on_stop
        if future_to_measure is not None:
            future_to_measure.add_done_callback(lambda fut: self.stop())

    @property
    def time_remaining(self) -> float:
        """
        :return: the difference between provided timeout and elapsed time, or None if timeout was
            not given. This will never be negative.
        """
        if self.timeout is None:
            return None
        v = self.timeout - self.get_time_elapsed()
        return 0 if v < 0 else v

    @property
    def timeouted(self) -> bool:
        """
        :return: Has the time elapsed exceeded timeout? Always False if timeout was not given
        """
        if self.timeout is None:
            return False
        return self.get_time_elapsed() >= self.timeout

    def assert_not_timeouted(self) -> None:
        """
        If the time elapsed exceeded timeout, throw WouldWaitMore.

        Always returns if the timeout was not givne
        """
        if self.timeout is None:
            return
        if self.timeouted:
            raise WouldWaitMore('timeout exceeded')

    def reset_and_start(self) -> None:
        """
        Syntactic sugar for calling reset() and then start()
        """
        self.reset()
        self.start()

    def has_exceeded(self, value: float) -> bool:
        """
        Return whether the timer has exceeded provided value.

        .. deprecated:: 2.14.22
        """
        warnings.warn('Use timeout parameter and timeouted property instead.',
                      PendingDeprecationWarning)
        return self() > value

    def raise_if_exceeded(self, value: float, exc_class: tp.Type[Exception] = WouldWaitMore):
        """
        Raise provided exception, with no arguments, if timer has clocked more than provided value.

        If no exc_class is provided, WouldWaitMore will be raised by default.

        .. deprecated:: 2.14.22
        """
        warnings.warn('Use timeout parameter and assert_not_timeouted property instead.',
                      PendingDeprecationWarning)
        if self.has_exceeded(value):
            raise exc_class()

    def reset(self) -> None:
        """
        Reset the counter, enabling it to start counting after a .stop() call.
        This will put the counter in a STOPPED mode if it's running already.
        """
        self.stopped_on = self.started_on = self.time_getter_callable()
        self.elapsed = 0

    def start(self) -> None:
        """Start measuring time or resume measuring it"""
        if not self.stop_on_stop:
            raise TypeError('stop_on_stop is disabled for this counter!')
        if self.stopped_on is None:
            raise TypeError('the counter is already running!')

        if self.stopped_on is not None:
            self.started_on = self.time_getter_callable() - self.elapsed
            self.stopped_on = None
        self.started_on = self.time_getter_callable()
        self.elapsed = None

    def update(self) -> None:
        """Alias for .start()"""
        self.start()

    def adjust(self, interval: float) -> None:
        """Add given value to internal started_at counter"""
        self.started_on += interval

    def get_time_elapsed(self) -> float:
        """
        :return: currently elapsed time
        """
        if self.stop_on_stop and self.elapsed is not None:
            return self.elapsed
        return self.time_getter_callable() - self.started_on

    def __call__(self, fun: tp.Optional[tp.Callable] = None) -> float:
        if fun is None:
            return self.get_time_elapsed()
        else:
            from satella.coding.decorators import auto_adapt_to_methods

            @auto_adapt_to_methods
            def outer(func):
                if inspect.isgeneratorfunction(func):
                    @wraps(func)
                    def inner(*args, **kwargs):
                        s = copy.copy(self)
                        if not s.create_stopped:
                            s.reset()
                            with s:
                                yield from func(s, *args, **kwargs)
                        else:
                            yield from func(s, *args, **kwargs)

                    return inner
                else:
                    @wraps(func)
                    def inner(*args, **kwargs):
                        s = copy.copy(self)
                        if not s.create_stopped:
                            s.reset()
                            with s:
                                return func(s, *args, **kwargs)
                        else:
                            return func(s, *args, **kwargs)

                    return inner

            return outer(fun)

    def __enter__(self):
        if self.elapsed is not None and self.stop_on_stop:
            self.start()
        return self

    def stop(self) -> None:
        """
        Stop counting time

        :raises TypeError: stop_on_stop is enabled or the counter has already been stopped
        """
        if not self.stop_on_stop:
            raise TypeError('stop_on_stop is disabled for this counter!')
        if self.stopped_on is not None:
            raise TypeError('counter already stopped!')

        self.stopped_on = self.time_getter_callable()
        self.elapsed = self.stopped_on - self.started_on

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        if self.stop_on_stop:
            self.stop()
        return False



import unittest

from satella.coding import ListDeleter, DictDeleter


class TestDeleters(unittest.TestCase):
    def test_dict_deleter(self):
        a = {1: 2, 2: 3, 3: 4, 4: 5, 5: 6}
        with DictDeleter(a) as ld:
            for key, value in ld.items():
                if key % 2 == 0:
                    ld.delete()

        self.assertEqual(a, {1: 2, 3: 4, 5: 6})

    def test_list_deleter_reversed(self):
        a = [1, 2, 3, 4, 5, 6, 7, 8]
        with ListDeleter(a) as ld:
            for v in reversed(ld):
                if v % 2:
                    ld.delete()
        self.assertEqual(a, [2, 4, 6, 8])

    def test_list_deleter(self):
        a = [1, 2, 3, 4, 5, 6, 7, 8]

        with ListDeleter(a) as ld:
            for v in ld:
                if v % 2 == 0:
                    ld.delete()

        self.assertEqual(a, [1, 3, 5, 7])



import os
import sys
import unittest

from satella.os.misc import suicide


class TestSuicide(unittest.TestCase):
    @unittest.skipIf('win' in sys.platform, 'Running on Windows')
    def test_suicide(self):
        pcid = os.fork()

        if pcid == 0:
            suicide()

        self.assertEqual(os.wait(), (pcid, 9))
